// AVL ...  

#include<stdio.h>
#include<stdlib.h>

//   
#define BALANCE_FLAG_R	1	//  
#define BALANCE_FLAG_H	2	//     

//! AVL 
class AVL_tree{
protected:
	//!  
	struct node_st {
		node_st *p1;	//!<  
		node_st *p2;	//!<  
		int value;		//!<  
		int b;			//!<  
		node_st(int v):p1(0),p2(0),b(0),value(v) {}
	};
	node_st *tree_root;	//!<  
private:
	bool Insert(node_st **node,int value);		//!<   ( )
	bool Remove(node_st **node,int value);		//!<  ( )
	bool GetMin(node_st **root,node_st **res);	//!<    ( )
	node_st *Rotate12(node_st*);				//!<  
	node_st *Rotate21(node_st*);				//!< 
	int Balance(node_st**,int);					//!<  
public:
	AVL_tree();
	~AVL_tree();
	void Insert(int value);		//!<  
	void Remove(int value);		//!<  
	bool Find(int value);		//!<  
};


AVL_tree::AVL_tree()
{
	tree_root=0; 
}

AVL_tree::~AVL_tree()
{
	node_st *t;
	if(!tree_root) return;
	//   
	while(tree_root->p1 || tree_root->p2) 
		if(tree_root->p2) t=tree_root->p2,tree_root->p2=t->p1,t->p1=tree_root,tree_root=t; //  
		else t=tree_root,tree_root=tree_root->p1,delete t;
	delete tree_root;
}


inline void AVL_tree::Insert(int value)
{
	Insert(&tree_root,value);
}

inline void AVL_tree::Remove(int value)
{
	Remove(&tree_root,value);
}

bool AVL_tree::Find(int value)
{
	node_st *node=tree_root;
	while(node) {
		if(node->value==value) return true;
		node=node->value>value?node->p1:node->p2;
	}
	return false;
}

int AVL_tree::Balance(node_st **root,int res=0)
{
	node_st *node=*root; 
	if(node->b>1) {
		res=BALANCE_FLAG_R;
		if(node->p2->b) res|=BALANCE_FLAG_H;
		if(node->p2->b<0) node->p2=Rotate12(node->p2);
		*root=Rotate21(node);
	}
	if(node->b<-1) {
		res=BALANCE_FLAG_R;
		if(node->p1->b) res|=BALANCE_FLAG_H;
		if(node->p1->b>0) node->p1=Rotate21(node->p1);
		*root=Rotate12(node);
	}
	return res;
}

bool AVL_tree::Insert(node_st **root,int value)
{
	bool res=false;
	node_st *node=*root;
	if(!node) {
		*root=new node_st(value);
		return true;
	}
	if(value==node->value) return false; 
	if(value<node->value) res=Insert(&node->p1,value) && !!--node->b;
	else                  res=Insert(&node->p2,value) && !!++node->b;
	if(Balance(root)&BALANCE_FLAG_R) res=false;
	return res;
}

AVL_tree::node_st *AVL_tree::Rotate12(node_st *node)
{
	static const int array[6][4]={
		-1,-1,+1,+1,
		-1,+0,+1,+0,
		-1,+1,+2,+0,
		-2,-1,+0,+0,
		-2,-2,+0,+1,
		-2,+0,+1,-1 
	};
	node_st *p1=node->p1;
	node_st *p12=p1->p2;
	p1->p2=node;
	node->p1=p12;
	for(int n=0;n<6;n++) if(array[n][0]==node->b && array[n][1]==p1->b) {
		p1->b=array[n][2];
		node->b=array[n][3];
		break;
	}
	return p1;
}

AVL_tree::node_st *AVL_tree::Rotate21(node_st *node)
{
	static const int array[6][4]={
		+1,-1,-2,+0,
		+1,+0,-1,+0,
		+1,+1,-1,-1,
		+2,+0,-1,+1,
		+2,+1,+0,+0,
		+2,+2,+0,-1 
	};
	node_st *p2=node->p2;
	node_st *p21=p2->p1;
	p2->p1=node;
	node->p2=p21;
	for(int n=0;n<6;n++) if(array[n][0]==node->b && array[n][1]==p2->b) {
		p2->b=array[n][2];
		node->b=array[n][3];
		break;
	}
	return p2;
}

// \retval res     
bool AVL_tree::GetMin(node_st **root,node_st **res)
{
	node_st *node=*root;
	if(node->p1) {
		if(GetMin(&node->p1,res) && !++node->b) return true;
		return !!(Balance(root)&BALANCE_FLAG_H);
	}
	*res=node;
	*root=node->p2;
	return true;
}


bool AVL_tree::Remove(node_st **root,int value)
{
	bool ok=false;
	node_st *node=*root;
	if(!node) return ok;
	if(node->value<value) {
		if(Remove(&node->p2,value) && !--node->b) ok=true;
	}else if(node->value>value) {
		if(Remove(&node->p1,value) && !++node->b) ok=true;
	}else {							//   .  
		if(!node->p2) {				//     , 
			*root=node->p1; delete node;
			return true;
		}
		ok=GetMin(&node->p2,root);	//  ,     
		(*root)->b =node->b;		//     ,  
		(*root)->p1=node->p1;
		(*root)->p2=node->p2;
		delete node;
		if(ok) ok=!--(*root)->b;
	} 
	return !!(Balance(root,ok?BALANCE_FLAG_H:0)&BALANCE_FLAG_H);
}




int main()
{
//	AVL_tree tree;
//	tree.Insert(1);
//	tree.Remove(2);
//	tree.Find(3);
	return 0;
}


